@@ -19,7 +19,7 @@
 ///        In-depth description of this algorithm:
 ///        https://github.com/kimwalisch/primecount/blob/master/doc/Easy-Special-Leaves.pdf
 ///
-/// Copyright (C) 2025 Kim Walisch, <kim.walisch@gmail.com>
+/// Copyright (C) 2026 Kim Walisch, <kim.walisch@gmail.com>
 ///
 /// This file is distributed under the BSD License. See the COPYING
 /// file in the top level directory.
@@ -86,6 +86,20 @@ T A(T x,
     sum += segmentedPi[xpq];
   }
 
+  // Unroll loop to increase instruction level parallelism
+  for (; i + 4 <= max_i2; i += 4)
+  {
+    uint64_t xpq0 = fast_div64(xp, primes[i]);
+    uint64_t xpq1 = fast_div64(xp, primes[i+1]);
+    uint64_t xpq2 = fast_div64(xp, primes[i+2]);
+    uint64_t xpq3 = fast_div64(xp, primes[i+3]);
+
+    sum += (segmentedPi[xpq0] * 2) +
+           (segmentedPi[xpq1] * 2) +
+           (segmentedPi[xpq2] * 2) +
+           (segmentedPi[xpq3] * 2);
+  }
+
   // pq = primes[b] * primes[i]
   // x / pq < y && low <= x / pq < high
   for (; i <= max_i2; i++)
@@ -178,8 +192,8 @@ T C2(T x,
   // many successive leaves are identical. If on average only
   // few successive leaves are identical then this loop
   // deteriorates performance due to poor instruction level
-  // parallelism and long instruction dependency chains.
-  if (avg_clustered_leaves >= 4 &&
+  // parallelism and increased cache misses.
+  if (avg_clustered_leaves >= 6 &&
       i > pi_min_clustered)
   {
     uint64_t iters = 0;
@@ -206,6 +220,20 @@ T C2(T x,
     avg_clustered_leaves = dist / iters;
   }
 
+  // Unroll loop to increase instruction level parallelism
+  for (; i >= pi_min_m + 4; i -= 4)
+  {
+    uint64_t xpq0 = fast_div64(xp, primes[i]);
+    uint64_t xpq1 = fast_div64(xp, primes[i-1]);
+    uint64_t xpq2 = fast_div64(xp, primes[i-2]);
+    uint64_t xpq3 = fast_div64(xp, primes[i-3]);
+
+    sum += (segmentedPi[xpq0] - b + 2) +
+           (segmentedPi[xpq1] - b + 2) +
+           (segmentedPi[xpq2] - b + 2) +
+           (segmentedPi[xpq3] - b + 2);
+  }
+
   // Find all sparse easy leaves where
   // successive leaves are different.
   // pq = primes[b] * primes[i]
@@ -389,6 +417,20 @@ T A_64(T xlow,
     sum += segmentedPi[xpq];
   }
 
+  // Unroll loop to increase instruction level parallelism
+  for (; i + 4 <= max_i2; i += 4)
+  {
+    uint64_t xpq0 = xp / primes[i];
+    uint64_t xpq1 = xp / primes[i+1];
+    uint64_t xpq2 = xp / primes[i+2];
+    uint64_t xpq3 = xp / primes[i+3];
+
+    sum += (segmentedPi[xpq0] * 2) +
+           (segmentedPi[xpq1] * 2) +
+           (segmentedPi[xpq2] * 2) +
+           (segmentedPi[xpq3] * 2);
+  }
+
   // pq = primes[b] * primes[i]
   // x / pq < y && low <= x / pq < high
   for (; i <= max_i2; i++)
@@ -433,6 +475,20 @@ T A_128(T xlow,
     sum += segmentedPi[xpq];
   }
 
+  // Unroll loop to increase instruction level parallelism
+  for (; i + 4 <= max_i2; i += 4)
+  {
+    uint64_t xpq0 = fast_div64(xp, primes[i]);
+    uint64_t xpq1 = fast_div64(xp, primes[i+1]);
+    uint64_t xpq2 = fast_div64(xp, primes[i+2]);
+    uint64_t xpq3 = fast_div64(xp, primes[i+3]);
+
+    sum += (segmentedPi[xpq0] * 2) +
+           (segmentedPi[xpq1] * 2) +
+           (segmentedPi[xpq2] * 2) +
+           (segmentedPi[xpq3] * 2);
+  }
+
   // pq = primes[b] * primes[i]
   // x / pq < y && low <= x / pq < high
   for (; i <= max_i2; i++)
@@ -525,8 +581,8 @@ T C2_64(T xlow,
   // many successive leaves are identical. If on average only
   // few successive leaves are identical then this loop
   // deteriorates performance due to poor instruction level
-  // parallelism and long instruction dependency chains.
-  if (avg_clustered_leaves >= 4 &&
+  // parallelism and increased cache misses.
+  if (avg_clustered_leaves >= 6 &&
       i > pi_min_clustered)
   {
     uint64_t iters = 0;
@@ -553,6 +609,20 @@ T C2_64(T xlow,
     avg_clustered_leaves = dist / iters;
   }
 
+  // Unroll loop to increase instruction level parallelism
+  for (; i >= pi_min_m + 4; i -= 4)
+  {
+    uint64_t xpq0 = xp / primes[i];
+    uint64_t xpq1 = xp / primes[i-1];
+    uint64_t xpq2 = xp / primes[i-2];
+    uint64_t xpq3 = xp / primes[i-3];
+
+    sum += (segmentedPi[xpq0] - b + 2) +
+           (segmentedPi[xpq1] - b + 2) +
+           (segmentedPi[xpq2] - b + 2) +
+           (segmentedPi[xpq3] - b + 2);
+  }
+
   // Find all sparse easy leaves where
   // successive leaves are different.
   // pq = primes[b] * primes[i]
@@ -601,8 +671,8 @@ T C2_128(T xlow,
   // many successive leaves are identical. If on average only
   // few successive leaves are identical then this loop
   // deteriorates performance due to poor instruction level
-  // parallelism and long instruction dependency chains.
-  if (avg_clustered_leaves >= 4 &&
+  // parallelism and increased cache misses.
+  if (avg_clustered_leaves >= 6 &&
       i > pi_min_clustered)
   {
     uint64_t iters = 0;
@@ -629,6 +699,20 @@ T C2_128(T xlow,
     avg_clustered_leaves = dist / iters;
   }
 
+  // Unroll loop to increase instruction level parallelism
+  for (; i >= pi_min_m + 4; i -= 4)
+  {
+    uint64_t xpq0 = fast_div64(xp, primes[i]);
+    uint64_t xpq1 = fast_div64(xp, primes[i-1]);
+    uint64_t xpq2 = fast_div64(xp, primes[i-2]);
+    uint64_t xpq3 = fast_div64(xp, primes[i-3]);
+
+    sum += (segmentedPi[xpq0] - b + 2) +
+           (segmentedPi[xpq1] - b + 2) +
+           (segmentedPi[xpq2] - b + 2) +
+           (segmentedPi[xpq3] - b + 2);
+  }
+
   // Find all sparse easy leaves where
   // successive leaves are different.
   // pq = primes[b] * primes[i]
